#include "TypeCheck.h"
#include "ParseError.h"
#include <stdexcept>
#include <iostream>
#include "Printer.h"
#include <climits>

void TypeCheck::visitClassDef(ClassDef *t) {
	const string name = t->getName()->string_;
	BindingPtr parent;
	if (t->getParent()) {
		parent = scope->classes[t->getParent()];
		if (!parent) {
			throw UndefinedError(t->getParent());
		}
	} else {
		parent = scope;
	}
	if (state == findClasses) {
		auto c = make_shared<ClassInfo>(t->getName(), parent);
		static_pointer_cast<ClassT>(c->type)->pident_->var = c;
		t->getName()->var = c;
		scope->classes << c;
		classDefs.emplace_back(c, t);
	}
}

void TypeCheck::visitPIdent(PIdent *p_ident)
{
	// visitString(p_ident->string_);
	auto var = scope->currentBinding->variables[p_ident];
	p_ident->var = var;
	lastType = var->type;
	if (auto ptr = dynamic_pointer_cast<ClassT>(lastType)) {
		if (!ptr->pident_ || !ptr->pident_->var.lock())
			throw runtime_error("either no pident or no var assigned found");
	}
}

void TypeCheck::visitProg(Prog *prog)
{
	prog->listtopdef_->accept(this);
}

void TypeCheck::visitFnDef(FnDef *fn_def)
{
	fn_def->fundef_->accept(this);
}

void TypeCheck::visitClDef(ClDef *cl_def)
{
	cl_def->classdef_->accept(this);
}

void TypeCheck::visitFuncDef(FuncDef *def)
{
	if (state <= findClasses) {
		return;
	}
	// findFunctions

	def->type_->accept(this);
	def->listarg_->accept(this);

	FunctionInfoPtr f = make_shared<FunctionInfo>(def, scope->currentClass);

	// pident nie trzeba bo definiujemy a nie odwołujemy się
	// FunctionInfo tworzy argumenty

	auto &target = getParentBinding()->functions;
	target << f;
}

void TypeCheck::visitClassDefN(ClassDefN *class_def_n)
{
	visitClassDef(class_def_n);
}

void TypeCheck::visitClassDefE(ClassDefE *class_def_e)
{
	visitClassDef(class_def_e);
}

void TypeCheck::visitClassBl(ClassBl *class_bl)
{
	/* Code For ClassBl Goes Here */

	class_bl->listclassblockdef_->accept(this);

}

void TypeCheck::visitClassMthd(ClassMthd *class_mthd)
{
	/* Code For ClassMthd Goes Here */

	class_mthd->fundef_->accept(this);

}

void TypeCheck::visitClassFld(ClassFld *decl)
{
	decl->type_->accept(this);
	decl->listitem_->accept(this);

	Type* type = decl->type_;

	for (Item * el : *decl->listitem_) {
		if (el->expr_) {
			throw ParseError("Initialization in class variable definition", decl);
		}

	VarInfoPtr var = make_shared<VarInfo>(el->pident_, type);
		el->pident_->var = var;
		scope->currentClass->variables << var;
	}

}

void TypeCheck::visitBlk(Blk *blk)
{
	BindingPtr binding = blk->binding.lock();
	if (!binding) {
		binding = make_shared<Binding>(scope->currentBinding);
		blk->binding = binding;
	}

	scope->currentBinding = binding;

	blk->liststmt_->accept(this);

	scope->currentBinding = binding->getParent();
}

void TypeCheck::visitBStmt(BStmt *b_stmt)
{
	b_stmt->block_->accept(this);
}

void TypeCheck::visitDecl(Decl *decl)
{
	Type* type = decl->type_;
	type->accept(this);

	if (dynamic_cast<Void*>(type)) {
		throw ParseError("Void variable", type);
	}

	for (Item * el : *decl->listitem_) {
		// el->accept(this); // nie trzeba bo el to tylko ident = expr
		if (el->expr_) {
			auto exprType = evalExpr(el->expr_);
			if (!type->isEqual(exprType, true)) {
				throw InvalidTypeError(*type, {exprType}, el->expr_);
			}
		}

		VarInfoPtr var = make_shared<VarInfo>(el->pident_, type->clone());
		el->pident_->var = var;
		scope->currentBinding->variables << var;
	}
}

void TypeCheck::visitAss(Ass *ass)
{
	auto a = evalExpr(ass->expr_1);
	auto b = evalExpr(ass->expr_2);
	if (!a->isEqual(b, true)) {
		throw InvalidTypeError(*a, {b}, ass->expr_1);
	}
}

void TypeCheck::visitIncr(Incr *incr)
{
	evalExpr<Int>(incr->expr_);
}

void TypeCheck::visitDecr(Decr *decr)
{
	evalExpr<Int>(decr->expr_);
}

void TypeCheck::visitRet(Ret *ret)
{
	auto type = evalExpr(ret->expr_);
	checkReturnStatement(type, ret);
}

void TypeCheck::visitVRet(VRet *v_ret)
{
	checkReturnStatement(make_shared<Void>(), v_ret);
}

void TypeCheck::visitCond(Cond *cond)
{
	auto type = evalExpr<Bool>(cond->expr_);
	cond->stmt_->accept(this);
	if (!dynamic_cast<ELitTrue*>(cond->expr_)) {
		returnType = nullptr;
	}
}

void TypeCheck::visitCondElse(CondElse *cond_else)
{
	evalExpr<Bool>(cond_else->expr_);
	cond_else->stmt_1->accept(this);
	auto ret1 = returnType;
	returnType = nullptr;
	cond_else->stmt_2->accept(this);
	auto ret2 = returnType;
	if (dynamic_cast<ELitTrue*>(cond_else->expr_)) {
		returnType = ret1;
	} else if (dynamic_cast<ELitFalse*>(cond_else->expr_)) {
		returnType = ret2;
	} else if (ret1 && ret2) {
		if (*ret1 != *ret2) {
			throw InvalidTypeError(*ret1, {ret2}, cond_else);
		}
	} else {
		returnType = nullptr;
	}

}

void TypeCheck::visitWhile(While *while_)
{
	evalExpr<Bool>(while_->expr_);
	while_->stmt_->accept(this);

	if (!dynamic_cast<ELitTrue*>(while_->expr_)) {
		returnType = nullptr;
	}
}

void TypeCheck::visitSExp(SExp *s_exp)
{
	evalExpr(s_exp->expr_);
}

void TypeCheck::visitForEach(ForEach *for_each)
{
	BindingPtr binding = make_shared<Binding>(scope->currentBinding);
	for_each->type_->accept(this);
	Array expect(for_each->type_);
	auto arrType = evalExpr<Array>(for_each->expr_);
	if (!expect.isEqual(&*arrType, true)) {
		throw InvalidTypeError(expect, {arrType}, for_each->expr_);
	}

	VarInfoPtr var = make_shared<VarInfo>(for_each->pident_, for_each->type_->clone());
	binding->variables << var;
	scope->currentBinding = binding;

	for_each->stmt_->accept(this);

	scope->currentBinding = binding->getParent();
}

void TypeCheck::visitInt(Int *int_)
{}

void TypeCheck::visitStr(Str *str)
{}

void TypeCheck::visitBool(Bool *bool_)
{}

void TypeCheck::visitVoid(Void *void_)
{}

void TypeCheck::visitArray(Array *array)
{
	array->binding = scope->currentBinding;
	array->type_->accept(this);
}

void TypeCheck::visitClassT(ClassT *class_t)
{
	class_t->binding = scope->currentBinding;
	ClassInfoPtr klass = scope->currentBinding->classes[class_t->pident_];
	class_t->pident_->var = klass;
}

void TypeCheck::visitFun(Fun *fun)
{
	fun->binding = scope->currentBinding;
	fun->type_->accept(this);
	fun->listtype_->accept(this);

}

void TypeCheck::visitEVar(EVar *e_var)
{
	e_var->pident_->accept(this);
	lastType = scope->currentBinding->variables[e_var->pident_]->type;
}

void TypeCheck::visitEIndexAcc(EIndexAcc *e_index_acc)
{
	auto type = evalExpr<Array>(e_index_acc->expr_1);
	evalExpr<Int>(e_index_acc->expr_2);

	shared_ptr<Type> last(type->type_->clone());
	last->binding = scope->currentBinding;
	lastType = last;
}

void TypeCheck::visitEClsMmbr(EClsMmbr *e_cls_mmbr)
{
	auto exprType = evalExpr(e_cls_mmbr->expr_);
	shared_ptr<ClassT> type = dynamic_pointer_cast<ClassT>(exprType);
	if (!type) {
		// it is an array and we ask for length
		if (dynamic_pointer_cast<Array>(exprType)) {
			if (e_cls_mmbr->pident_->string_ == "length") {
				lastType = make_shared<Int>();
				return;
			}
		}
		throw ParseError("Class expected", e_cls_mmbr->expr_);
	}
	ClassInfoPtr klass = scope->classes[type->pident_];
	if (klass != type->pident_->var.lock()) {
		throw ParseError("ej lol inny var");
	}
	VarInfoPtr var = klass->variables[e_cls_mmbr->pident_];
	if (!var || var == scope->variables.local(e_cls_mmbr->pident_->string_)) {
		cout << "undef clsmmbr !var" << endl;
		throw UndefinedError(e_cls_mmbr->pident_);
	}
	lastType = var->type;
}

void TypeCheck::visitEApp(EApp *e_app)
{
	shared_ptr<Fun> type = evalExpr<Fun>(e_app->expr_);
	if (type->listtype_->size() != e_app->listexpr_->size()) {
		string msg = "Expected " + to_string(type->listtype_->size()) + " arguments, " + to_string(e_app->listexpr_->size()) + " given";
		throw ParseError(msg, e_app);
	}
	for (unsigned int i = 0; i < type->listtype_->size(); i++) {
		try {
			Expr *expr = e_app->listexpr_->at(i);
			auto exprType = evalExpr(expr);
			auto *expect = type->listtype_->at(i);
			if (!expect->isEqual(exprType, true)) {
				throw InvalidTypeError(*expect, {exprType}, expr);
			}
		} catch (const ParseError &e) {
			string msg = string(e.what()) + "\nIn argument " + to_string(i+1) + " of function call:";
			ParseError ctx(msg, e_app);
			ctx.line = e.line;
			throw e;
		}
	}

	lastType = shared_ptr<Type>(type->type_->clone());
}

void TypeCheck::visitELitInt(ELitInt *e_lit_int)
{
	visitInteger(e_lit_int->integer_);
	if (e_lit_int->integer_ > INT_MAX || e_lit_int->integer_ < INT_MIN) {
		throw ParseError("constant too big", e_lit_int->lineno);
	}
	lastType = make_shared<Int>();
}

void TypeCheck::visitELitTrue(ELitTrue *e_lit_true)
{
	lastType = make_shared<Bool>();
}

void TypeCheck::visitELitFalse(ELitFalse *e_lit_false)
{
	lastType = make_shared<Bool>();
}

void TypeCheck::visitEString(EString *e)
{
	visitString(e->string_);
	lastType = make_shared<Str>();
}

void TypeCheck::visitENewArray(ENewArray *e)
{
	evalExpr<Int>(e->expr_);
	e->type_->accept(this);
	if (dynamic_cast<Void*>(e->type_)) {
		throw ParseError("Void table", e);
	}
	auto ret = make_shared<Array>(e->type_->clone()); // make type
	ret->type_->binding = scope->currentBinding;
	lastType = ret;
}

void TypeCheck::visitENewClass(ENewClass *e_new_class)
{
	// e_new_class->pident_->accept(this);
	lastType = evalExpr<ClassT>(e_new_class->pident_);
	// if (!e_new_class->pident_->var.lock()) {

	// }
	// ClassInfoPtr klass = dynamic_pointer_cast<ClassInfo>(e_new_class->pident_->var.lock());
	// if (!klass) {
	// 	throw ParseError("Class expected", e_new_class->pident_);
	// }

	// auto ret = make_shared<ClassT>(e_new_class->pident_); // make type
	// ret->binding = scope->currentBinding;
	// lastType = ret;
}

void TypeCheck::visitNullCast(NullCast *null_cast)
{
	lastType = evalExpr<ClassT>(null_cast->expr_);
}

void TypeCheck::visitNeg(Neg *e)
{
	lastType = evalExpr<Int>(e->expr_);
}

void TypeCheck::visitNot(Not *e)
{
	lastType = evalExpr<Bool>(e->expr_);
}

void TypeCheck::visitEMul(EMul *e)
{
	auto a = evalExpr<Int>(e->expr_1);
	auto b = evalExpr<Int>(e->expr_2);
	lastType = b;
}

void TypeCheck::visitEAdd(EAdd *e)
{
	auto a = evalExpr(e->expr_1);
	auto b = evalExpr(e->expr_2);
	if (dynamic_cast<Plus*>(e->addop_)) {
		if (*a != *b) {
			throw InvalidTypeError(*a, {b}, e);
		}
	} else {
		Int expect;
		if (*a != expect || *a != *b) {
			throw InvalidTypeError(expect, {a, b}, e);
		}
	}
	lastType = b;
}

void TypeCheck::visitERel(ERel *e)
{
	auto a = evalExpr(e->expr_1);
	auto b = evalExpr(e->expr_2);
	if (dynamic_cast<EQU*>(e->relop_) || dynamic_cast<NE*>(e->relop_)) {
		if (!a->isEqual(b, true) && !b->isEqual(a, true)) {
			throw InvalidTypeError(*a, {b}, e);
		}
	} else {
		Int expect;
		if (*a != expect || *a != *b) {
			throw InvalidTypeError(expect, {a, b}, e);
		}
	}
	lastType = make_shared<Bool>();
}

void TypeCheck::visitEAnd(EAnd *e)
{
	auto a = evalExpr<Bool>(e->expr_1);
	auto b = evalExpr<Bool>(e->expr_2);
	lastType = b;
}

void TypeCheck::visitEOr(EOr *e)
{
	auto a = evalExpr<Bool>(e->expr_1);
	auto b = evalExpr<Bool>(e->expr_2);
	lastType = b;
}

void TypeCheck::visitListTopDef(ListTopDef *list_top_def)
{
	for (auto &i : *list_top_def)
	{
		i->accept(this);
	}
}

void TypeCheck::visitListArg(ListArg *list_arg)
{
	for (ListArg::iterator i = list_arg->begin() ; i != list_arg->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListClassBlockDef(ListClassBlockDef *list_class_block_def)
{
	for (ListClassBlockDef::iterator i = list_class_block_def->begin() ; i != list_class_block_def->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListStmt(ListStmt *list_stmt)
{
	for (ListStmt::iterator i = list_stmt->begin() ; i != list_stmt->end() ; ++i)
	{
		(*i)->accept(this);
		if (returnType) break;
	}
}

void TypeCheck::visitListItem(ListItem *list_item)
{
	for (ListItem::iterator i = list_item->begin() ; i != list_item->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListType(ListType *list_type)
{
	for (ListType::iterator i = list_type->begin() ; i != list_type->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListExpr(ListExpr *list_expr)
{
	for (ListExpr::iterator i = list_expr->begin() ; i != list_expr->end() ; ++i)
	{
		(*i)->accept(this);
	}
}


void TypeCheck::visitInteger(Integer x) {}
void TypeCheck::visitString(String x) {}


TypeCheck::TypeCheck()
	: state(initialized), scope(make_shared<Scope>())
{}

void TypeCheck::addIOFunctions(Type &type) {
	string name = type.printName();
	name[0] += 'A' - 'a';
	TypePtr fun = make_shared<Fun>(type.clone(), new ListType()); // read
	FunctionInfoPtr read = make_shared<FunctionInfo>("read" + name, fun);

	scope->functions << read;

	auto *v = new ListType();
	v->push_back(type.clone());
	fun = make_shared<Fun>(new Void(), v); // print
	FunctionInfoPtr print = make_shared<FunctionInfo>("print" + name, fun);

	scope->functions << print;

}
void TypeCheck::setupEnv() {
	{ Int x; addIOFunctions(x); }
	{ Str x; addIOFunctions(x); }
	// { Bool x; addIOFunctions(x); }

	// void error()
	{
		TypePtr fun = make_shared<Fun>(Void().clone(), new ListType());
		FunctionInfoPtr error = make_shared<FunctionInfo>("error", fun);

		scope->functions << error;
	}
}

void TypeCheck::check(Visitable *v)
{
	if (state != State::initialized) {
		throw std::runtime_error("already initialized");
	}
	setupEnv();
	state = State::findClasses;

	scope->currentBinding = scope;

	// najpierw zebranie nazw klas
	v->accept(this);

	state = State::findFunctions;
	scope->currentBinding = scope;
	// zebranie funkcji zewnętrznych
	v->accept(this);

	state = State::visit;

	// teraz bloki klas jak juz znamy wszystkie istniejace klasy i funkcje
	for (auto p : classDefs) {
		scope->currentBinding = scope->currentClass = p.first;
		p.second->getBlock()->accept(this);
		scope->currentClass = nullptr;
		scope->currentBinding = scope;
	}
	classDefs.clear();


	// a teraz mozna funkcje w klasach
	for (auto c : scope->classes) {
		scope->currentClass = c;
		try {
			for (auto f : c->functions) {
				checkFunction(f);
			}
		} catch (const ParseError &e) {
			throw ParseError(e, *c);
		}


		scope->currentClass = nullptr;
	}
	for (auto f : scope->functions) {
		checkFunction(f);
	}

	if (FunctionInfoPtr main = scope->functions.local("main")) {
		Fun expected(new Int(), new ListType());
		if (!main->type->isEqual(&expected, false)) {
			throw InvalidTypeError(expected, {main->type}, main->ident);
		}
	} else {
		throw ParseError("No \"main\" function declared");
	}

}

void TypeCheck::checkReturnStatement(shared_ptr<Type> type, Stmt *stmt) {
	lastType = returnType = type;
	auto f = scope->currentFunction;
	auto funType = dynamic_pointer_cast<Fun>(f->type);
	if (!funType) throw runtime_error("FunctionInfo " + f->name + ":" + to_string(f->lineLocation) + " nie ma typu Fun!");
	auto funRet = funType->type_;

	if (!funRet->isEqual(type, true)) {
		throw InvalidTypeError(*funRet, {type}, stmt);
	}
}


void TypeCheck::checkFunction(FunctionInfoPtr f)
{
	if (!f->block) return;
	try {
		scope->currentFunction = f;

		if (f->binding.lock()) throw runtime_error("Ale ten binding juz istnieje");

		BindingPtr binding = make_shared<Binding>(getParentBinding());
		f->binding = binding;
		if (scope->currentClass) {
			binding->variables << make_shared<VarInfo>("self", scope->currentClass->type);
		}

		for (auto arg : f->arguments) {
			binding->variables << arg;
		}

		scope->currentBinding = binding;
		returnType = lastType = nullptr;

		f->block->accept(this);

		checkFunctionRet();

		scope->currentBinding = nullptr;

		scope->currentFunction = nullptr;
	} catch (const ParseError &e) {
		throw ParseError(e, *f);
	}
}

void TypeCheck::checkFunctionRet() {
	auto &f = scope->currentFunction;
	auto funType = dynamic_pointer_cast<Fun>(f->type);
	if (!funType) throw runtime_error("FunctionInfo " + f->name + ":" + to_string(f->lineLocation) + " nie ma typu Fun!");
	auto funRet = funType->type_;
	if (*funRet != Void()) {
		if (!returnType || returnType == nullptr) {
			throw ParseError("Function " + f->name + " did not return!", f->lineLocation);
		}
		if (!funRet->isEqual(returnType, true)) {
			throw InvalidTypeError(*returnType, {funRet}, f->block);
		}
	} else {
		if (returnType && !funRet->isEqual(returnType, true)) {
			throw InvalidTypeError(*returnType, {funRet}, f->block);
		}
	}
}
